#!/usr/bin/env python
#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Softwar where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import os, sys, shutil

# Find common code relative in file system. Not using packages for test stuff.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(
            os.path.dirname( os.path.abspath( __file__ ) ),
            ".."
        )
    )
)
from test_common import (
    my_print,
    setup,
    compareWithCPython,
    decideFilenameVersionSkip,
    getRuntimeTraceOfLoadedFiles
)

python_version = setup()

search_mode = len( sys.argv ) > 1 and sys.argv[1] == "search"

start_at = sys.argv[2] if len( sys.argv ) > 2 else None

if start_at:
    active = False
else:
    active = True

for filename in sorted( os.listdir( "." ) ):
    if not filename.endswith( ".py" ):
        continue

    if not decideFilenameVersionSkip( filename ):
        continue

    path = os.path.relpath( filename )

    if not active and start_at in ( filename, path ):
        active = True

    extra_flags = [ "expect_success", "standalone", "remove_output" ]

    if active:
        my_print( "Consider output of recursively compiled program:", filename )

        # First compare so we know the program behaves identical.
        compareWithCPython(
            path        = filename,
            extra_flags = extra_flags,
            search_mode = search_mode,
            needs_2to3  = False
        )

        # Second use strace on the result.
        loaded_filenames = getRuntimeTraceOfLoadedFiles(
            path = os.path.join(
                filename[:-3] + ".dist",
                filename[:-3] + ".exe"
            )
        )

        shutil.rmtree(filename[:-3] + ".dist")

        current_dir = os.path.normpath(os.getcwd())

        for loaded_filename in loaded_filenames:
            loaded_filename = os.path.normpath(loaded_filename)

            if loaded_filename.startswith(current_dir):
                continue

            if loaded_filename.startswith("/etc/"):
                continue

            if loaded_filename.startswith("/proc/"):
                continue

            if loaded_filename.startswith("/dev/"):
                continue

            if loaded_filename.startswith( "/usr/lib/locale/" ):
                continue

            if loaded_filename.startswith( "/usr/share/locale/" ):
                continue

            if loaded_filename.startswith("/lib/libdl.") or \
               loaded_filename.startswith("/lib64/libdl."):
                continue

            if loaded_filename.startswith("/lib/libm.") or \
               loaded_filename.startswith("/lib64/libm."):
                continue

            if loaded_filename.startswith("/lib/libutil.") or \
               loaded_filename.startswith("/lib64/libutil."):
                continue

            if loaded_filename.startswith("/lib/libpthread.") or \
               loaded_filename.startswith("/lib64/libpthread."):
                continue

            if os.path.basename(loaded_filename) == "gconv-modules.cache":
                continue

            sys.exit("Should not access '%s'." % loaded_filename)
    else:
        my_print( "Skipping", filename )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Blob codes for storing binary data semi-efficiently.

This module offers means to store and encode binary blobs in C++ semi
efficiently. The "StreamData" class is used in two places, for constants
and for freezing of bytecode.
"""

from nuitka import Utils

class StreamData:
    def __init__( self ):
        self.stream_data = bytes()

    def encodeStreamData( self ):
        for count, stream_byte in enumerate( self.stream_data ):
            if count % 16 == 0:
                if count > 0:
                    yield "\n"
                yield "   "

            if Utils.python_version < 300:
                yield " 0x%02x," % ord( stream_byte )
            else:
                yield " 0x%02x," % stream_byte

    def getStreamDataCode( self, value, fixed_size = False ):
        offset = self.stream_data.find( value )
        if offset == -1:
            offset = len( self.stream_data )
            self.stream_data += value

        if fixed_size:
            return "&stream_data[ %d ]" % offset
        else:
            return "&stream_data[ %d ], %d" % ( offset, len( value ) )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Templates for the constants handling.

"""

template_constants_reading = """
#include "nuitka/prelude.hpp"

// Sentinel PyObject to be used for all our call iterator endings. It will
// become a PyCObject pointing to NULL. It's address is unique, and that's
// enough for us to use it as sentinel value.
PyObject *_sentinel_value = NULL;

%(constant_declarations)s

// Blob from which objects are unstreamed.
static const unsigned char stream_data[] =
{
%(stream_data)s
};

static void __initConstants( void )
{
%(constant_inits)s
}

void _initConstants( void )
{
    if ( _sentinel_value == NULL )
    {
#if PYTHON_VERSION < 300
        _sentinel_value = PyCObject_FromVoidPtr( NULL, NULL );
#else
        // The NULL value is not allowed for a capsule, so use something else.
        _sentinel_value = PyCapsule_New( (void *)27, "sentinel", NULL );
#endif
        assert( _sentinel_value );

        __initConstants();
    }
}
"""

template_constants_declaration = """\
// Call this to initialize all of the below
void _initConstants( void );

%(constant_declarations)s
"""



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from __future__ import print_function

import os, sys, subprocess, tempfile, atexit, shutil, re

# Make sure we flush after every print, the "-u" option does more than that
# and this is easy enough.
def my_print( *args, **kwargs ):
    print( *args, **kwargs )
    sys.stdout.flush()

def check_output(*popenargs, **kwargs):
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')

    process = subprocess.Popen( stdout = subprocess.PIPE, *popenargs, **kwargs )
    output, _unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise subprocess.CalledProcessError( retcode, cmd, output=output )
    return output

def setup( needs_io_encoding = False ):
    # Go its own directory, to have it easy with path knowledge.
    os.chdir(
        os.path.dirname(
            os.path.abspath( sys.modules[ "__main__" ].__file__ )
        )
    )

    if "PYTHON" not in os.environ:
        os.environ[ "PYTHON" ] = sys.executable

    if needs_io_encoding and "PYTHONIOENCODING" not in os.environ:
        os.environ[ "PYTHONIOENCODING" ] = "utf-8"

    version_output = check_output(
        [ os.environ[ "PYTHON" ], "--version" ],
        stderr = subprocess.STDOUT
    )

    global python_version
    python_version = version_output.split()[1]

    my_print( "Using concrete python", python_version )

    assert type( python_version ) is bytes

    return python_version

tmp_dir = None

def getTempDir():
    # Create a temporary directory to work in, automatically remove it in case
    # it is empty in the end.
    global tmp_dir

    if tmp_dir is None:
        tmp_dir = tempfile.mkdtemp(
            prefix = os.path.basename(
                os.path.dirname(
                    os.path.abspath( sys.modules[ "__main__" ].__file__ )
                )
            ) + "-",
            dir    = tempfile.gettempdir() if
                         not os.path.exists( "/var/tmp" ) else
                    "/var/tmp"
        )

        def removeTempDir():
            try:
                os.rmdir( tmp_dir )
            except OSError:
                pass

        atexit.register( removeTempDir )

    return tmp_dir

def convertUsing2to3( path ):
    filename = os.path.basename( path )

    new_path = os.path.join( getTempDir(), filename )
    shutil.copy( path, new_path )

    # On Windows, we cannot rely on 2to3 to be in the path.
    if os.name == "nt":
        command = [
            sys.executable,
            os.path.join(
                os.path.dirname( sys.executable ),
                "Tools/Scripts/2to3.py"
            )
        ]
    else:
        command = [ "2to3" ]

    command += [
        "-w",
        "-n",
        "--no-diffs",
        new_path
    ]

    check_output(
        command,
        stderr = open( os.devnull, "w" ),
    )

    return new_path

def decideFilenameVersionSkip( filename ):
    if filename.startswith( "run_" ):
        return False

    # Skip tests that require Python 2.7 at least.
    if filename.endswith( "27.py" ) and python_version.startswith( b"2.6" ):
        return False

    if filename.endswith( "_2.py" ) and python_version.startswith( b"3" ):
        return False

    # Skip tests that require Python 3.2 at least.
    if filename.endswith( "32.py" ) and not python_version.startswith( b"3" ):
        return False

    # Skip tests that require Python 3.3 at least.
    if filename.endswith( "33.py" ) and not python_version.startswith( b"3.3" ):
        return False

    return True

def compareWithCPython( path, extra_flags, search_mode, needs_2to3 ):
    # Apply 2to3 conversion if necessary.
    if needs_2to3:
        path = convertUsing2to3( path )

    command = [
        sys.executable,
        os.path.join( "..", "..", "bin", "compare_with_cpython" ),
        path,
        "silent"
    ]
    command += extra_flags

    try:
        result = subprocess.call(
            command
        )
    except KeyboardInterrupt:
        result = 2

    if result != 0 and result != 2 and search_mode:
        my_print("Error exit!", result)
        sys.exit( result )

    if needs_2to3:
        os.unlink( path )

    if result == 2:
        sys.stderr.write( "Interruped, with CTRL-C\n" )
        sys.exit( 2 )

def hasDebugPython():
    global python_version

    # On Debian systems, these work.
    debug_python = os.path.join( "/usr/bin/", os.environ[ "PYTHON" ] + "-dbg" )
    if os.path.exists( debug_python ):
        return True

    # For self compiled Python, if it's the one also executing the runner, lets
    # use it.
    if sys.executable == os.environ[ "PYTHON" ] and \
       hasattr( sys, "gettotalrefcount" ):
        return True

    # Otherwise no.
    return False

def getRuntimeTraceOfLoadedFiles( path ):
    """ Returns the files loaded when executing a binary. """

    result = []

    if os.name == "posix":
        args = (
            "strace",
            "-e", "file",
            "-s4096", # Some paths are truncated otherwise.
            path
        )

        process = subprocess.Popen(
            args   = args,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE
        )

        stdout_strace, stderr_strace = process.communicate()

        for line in stderr_strace.split("\n"):
            if not line:
                continue

            result.extend(
                os.path.abspath(match)
                for match in
                re.findall('"(.*?)"', line)
            )
    # TODO: Use depends.exe on Windows.

    result = list(sorted(set(result)))

    return result



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Pack and copy files for standalone mode.

This is in heavy flux now, cannot be expected to work or make sense on all
the platforms.
"""

import os
import subprocess
import sys
from logging import debug
import marshal

from nuitka import Utils
from nuitka.codegen.ConstantCodes import needsPickleInit


python_dll_dir_name = "_python"


def loadCodeObjectData(precompiled_path):
    # Ignoring magic numbers, etc. which we don't have to care for much as
    # CPython already checked them (would have rejected it otherwise).
    return open(precompiled_path, "rb").read()[8:]


def detectEarlyImports():
    command = "import encodings.utf_8;"

    if Utils.python_version < 300:
        command += "import encodings.hex_codec;"

    # When we are using pickle internally (for some hard constant cases we do),
    # we need to make sure it will be available as well.
    if needsPickleInit():
        command += "import {pickle};".format(
            pickle = "pickle" if Utils.python_version >= 300 else "cPickle"
        )

    # For Python3 we patch inspect without knowing if it is used.
    if Utils.python_version >= 300:
        command += "import inspect;"

    # Print statements for stuff to show.
    if Utils.python_version >= 300:
        command += r'import sys; print("\n".join(sorted("import " + module.__name__ + " # sourcefile " + module.__file__ for module in sys.modules.values() if hasattr(module, "__file__") and module.__file__ != "<frozen>")), file = sys.stderr)'  # do not read it, pylint: disable=C0301  lint:ok

    process = subprocess.Popen(
        args   = [sys.executable, "-s", "-S", "-v", "-c", command],
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    _stdout, stderr = process.communicate()

    result = []

    debug("Detecting early imports:")

    # bug of PyLint, pylint: disable=E1103
    for line in stderr.replace(b"\r", b"").split(b"\n"):
        if line.startswith(b"import "):
            # print(line)

            parts = line.split(b" # ", 2)

            module_name = parts[0].split(b" ", 2)[1]
            origin = parts[1].split()[0]

            if origin == b"precompiled":
                # This is a ".pyc" file that was imported, even before we have a
                # chance to do anything, we need to preserve it.
                filename = parts[1][len(b"precompiled from "):]

                debug(
                    "Freezing module '%s' (from '%s').",
                    module_name,
                    filename
                )

                result.append(
                    (
                        module_name,
                        loadCodeObjectData( filename ),
                        b"__init__" in filename
                    )
                )

            elif origin == b"sourcefile":
                filename = parts[1][len(b"sourcefile "):]

                debug(
                    "Freezing module '%s' (from '%s').",
                    module_name,
                    filename
                )

                source_code = open(filename,"rb").read()

                if Utils.python_version >= 300:
                    source_code = source_code.decode( "utf-8" )

                result.append(
                    (
                        module_name,
                        marshal.dumps(
                            compile(source_code, filename, "exec")
                        ),
                        Utils.basename(filename) == b"__init__.py"
                    )
                )

    debug("Finished detecting early imports.")

    return result

def _detectPythonDLLs( binary_filename ):
    result = set()

    if os.name == "posix" and os.uname()[0] == "Linux":
        # Ask "ldd" about the libraries being used by the created binary, these
        # are the ones that interest us.
        process = subprocess.Popen(
            args   = [
                "ldd",
                binary_filename
            ],
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE
        )

        stdout, _stderr = process.communicate()

        for line in stdout.split(b"\n"):
            if not line:
                continue

            if b"=>" not in line:
                continue

            part = line.split(b" => ", 2)[1]
            filename = part[:part.rfind(b"(")-1]

            if not filename:
                continue

            if Utils.python_version >= 300:
                filename = filename.decode("utf-8")

            result.add(filename)
    elif os.name == "nt":
        import ctypes
        from ctypes import windll
        from ctypes.wintypes import HANDLE, LPCSTR, DWORD
        dll = getattr( windll, "python%s%s" % ( sys.version_info[:2] ) )
        getname = windll.kernel32.GetModuleFileNameA
        getname.argtypes = ( HANDLE, LPCSTR, DWORD )
        getname.restype = DWORD
        c_result = ctypes.create_string_buffer( 1024 )
        size = getname(dll._handle,c_result, 1024 )
        path = c_result.value[:size]

        if Utils.python_version >= 300:
            path = path.decode("utf-8")

        result.add(path)
    else:
        # Support your platform above.
        assert False

    return result


def detectPythonDLLs( standalone_entry_points ):
    result = set()

    for binary_filename in standalone_entry_points:
        result.update(
            _detectPythonDLLs(binary_filename)
        )

    return result



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Module/Package nodes

The top of the tree. Packages are also modules. Modules are what hold a program
together and cross-module optimizations are the most difficult to tackle.
"""

from .NodeBases import (
    ClosureGiverNodeBase,
    ChildrenHavingMixin,
    NodeBase
)

from .IndicatorMixins import MarkContainsTryExceptIndicator
from nuitka.SourceCodeReferences import SourceCodeReference
from nuitka.nodes.FutureSpecs import FutureSpec

from nuitka import Variables, Utils

from nuitka.oset import OrderedSet

class PythonModuleMixin:
    def __init__( self, name, package_name ):
        assert type(name) is str, type(name)
        assert "." not in name, name
        assert package_name is None or \
               ( type( package_name ) is str and package_name != "" )

        self.name = name
        self.package_name = package_name
        self.package = None

    def getName( self ):
        return self.name

    def getPackage( self ):
        return self.package_name

    def getFullName( self ):
        if self.package_name:
            return self.package_name + "." + self.getName()
        else:
            return self.getName()

    def isMainModule( self ):
        return False

    def isInternalModule( self ):
        return False


class PythonModule( PythonModuleMixin, ChildrenHavingMixin,
                    ClosureGiverNodeBase, MarkContainsTryExceptIndicator ):
    """ Module

        The module is the only possible root of a tree. When there are many
        modules they form a forrest.
    """

    kind = "PYTHON_MODULE"

    named_children = ( "body", )

    def __init__( self, name, package_name, source_ref ):

        ClosureGiverNodeBase.__init__(
            self,
            name        = name,
            code_prefix = "module",
            source_ref  = source_ref
        )

        ChildrenHavingMixin.__init__(
            self,
            values = {}
        )

        MarkContainsTryExceptIndicator.__init__( self )

        PythonModuleMixin.__init__(
            self,
            name         = name,
            package_name = package_name
        )

        self.variables = set()

        # The list functions contained in that module.
        self.functions = OrderedSet()

        self.active_functions = OrderedSet()

        # SSA trace based information about the module.
        self.collection = None

    def getDetails( self ):
        return {
            "filename" : self.source_ref.getFilename(),
            "package"  : self.package_name,
            "name"     : self.name
        }

    def asXml( self ):
        # The class is new style, false alarm: pylint: disable=E1002
        result = super( PythonModule, self ).asXml()

        for function_body in self.functions:
            result.append( function_body.asXml() )

        return result

    getBody = ChildrenHavingMixin.childGetter( "body" )
    setBody = ChildrenHavingMixin.childSetter( "body" )

    def isPythonModule( self ):
        return True

    def getParent( self ):
        assert False

    def getParentVariableProvider( self ):
        return None

    def getVariables( self ):
        return self.variables

    def getFilename( self ):
        return self.source_ref.getFilename()

    def getVariableForAssignment( self, variable_name ):
        result = self.getProvidedVariable( variable_name )

        return result.makeReference( self )

    def getVariableForReference( self, variable_name ):
        result = self.getProvidedVariable( variable_name )

        return result.makeReference( self )

    def getVariableForClosure( self, variable_name ):
        return self.getProvidedVariable(
            variable_name = variable_name
        )

    def createProvidedVariable( self, variable_name ):
        result = Variables.ModuleVariable(
            module        = self,
            variable_name = variable_name
        )

        assert result not in self.variables
        self.variables.add( result )

        return result

    def isEarlyClosure( self ):
        # Modules should immediately closure variables on use.
        # pylint: disable=R0201
        return True

    def getCodeName( self ):
        return "module_" + self.getFullName().replace( ".", "__" ).replace( "-", "_" )

    def addFunction( self, function_body ):
        assert function_body not in self.functions

        self.functions.add( function_body )

    def getFunctions( self ):
        return self.functions

    def startTraversal( self ):
        self.active_functions = OrderedSet()

    def addUsedFunction( self, function_body ):
        assert function_body in self.functions

        assert function_body.isExpressionFunctionBody()

        if function_body not in self.active_functions:
            self.active_functions.add( function_body )

    def getUsedFunctions( self ):
        return self.active_functions

    def getOutputFilename( self ):
        main_filename = self.getFilename()

        if main_filename.endswith( ".py" ):
            return main_filename[:-3]
        else:
            return main_filename

    def attemptRecursion( self, constraint_collection ):
        # Make sure the package is recursed to.
        from nuitka.tree import Recursion
        from nuitka import Importing

        if self.package_name is not None and self.package is None:
            package_package, _package_module_name, package_filename = \
              Importing.findModule(
                source_ref     = self.getSourceReference(),
                module_name    = self.package_name,
                parent_package = None,
                level          = 1,
                warn           = Utils.python_version < 330
            )

            # TODO: Temporary, if we can't find the package for Python3.3 that
            # is semi-OK, maybe.
            if Utils.python_version >= 330 and not package_filename:
                return

            imported_module, is_added = Recursion.recurseTo(
                module_package  = package_package,
                module_filename = package_filename,
                module_relpath  = Utils.relpath(package_filename),
                module_kind     = "py",
                reason          = "Containing package of recursed module.",
            )

            self.package = imported_module

            if is_added:
                constraint_collection.signalChange(
                    "new_code",
                    imported_module.getSourceReference(),
                    "Recursed to module package."
                )

        if self.package:
            from nuitka.ModuleRegistry import addUsedModule

            addUsedModule( self.package )

#            print "Recursed to package", self.package_name
            self.package.attemptRecursion( constraint_collection )


class SingleCreationMixin:
    created = set()

    def __init__( self ):
        assert self.__class__ not in self.created
        self.created.add( self.__class__ )


class PythonMainModule( PythonModule, SingleCreationMixin ):
    kind = "PYTHON_MAIN_MODULE"

    def __init__( self, main_added, source_ref ):
        PythonModule.__init__(
            self,
            name         = "__main__",
            package_name = None,
            source_ref   = source_ref
        )

        SingleCreationMixin.__init__( self )

        self.main_added = main_added

    def isMainModule( self ):
        return True

    def getOutputFilename( self ):
        if self.main_added:
            return Utils.dirname( self.getFilename() )
        else:
            return PythonModule.getOutputFilename( self )


class PythonInternalModule( PythonModule, SingleCreationMixin ):
    kind = "PYTHON_INTERNAL_MODULE"

    def __init__( self ):
        PythonModule.__init__(
            self,
            name         = "__internal__",
            package_name = None,
            source_ref   = SourceCodeReference.fromFilenameAndLine(
                filename    = "internal",
                line        = 0,
                future_spec = FutureSpec(),
                inside_exec = False
            )
        )

        SingleCreationMixin.__init__( self )

    def isInternalModule( self ):
        return True

    def getOutputFilename( self ):
        return "__internal"


class PythonPackage( PythonModule ):
    kind = "PYTHON_PACKAGE"

    def __init__( self, name, package_name, source_ref ):
        assert name

        PythonModule.__init__(
            self,
            name         = name,
            package_name = package_name,
            source_ref   = source_ref
        )

    def getOutputFilename( self ):
        return Utils.dirname( self.getFilename() )


class PythonShlibModule( PythonModuleMixin, NodeBase ):
    kind = "PYTHON_SHLIB_MODULE"

    def __init__( self, name, package_name, source_ref ):
        NodeBase.__init__( self, source_ref )

        PythonModuleMixin.__init__(
            self,
            name         = name,
            package_name = package_name
        )

    def getDetails( self ):
        return {
            "name": self.name,
            "package_name" : self.package_name
        }

    def getFilename( self ):
        return self.getSourceReference().getFilename()

    def startTraversal( self ):
        pass



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Code templates one stop access. """

# Wildcard imports are here to centralize the templates for access through one module
# name, this one, they are not used here though. pylint: disable=W0401,W0614

from .templates.CodeTemplatesMain import *
from .templates.CodeTemplatesConstants import *

from .templates.CodeTemplatesFunction import *
from .templates.CodeTemplatesGeneratorFunction import *

from .templates.CodeTemplatesParameterParsing import *

from .templates.CodeTemplatesExceptions import *
from .templates.CodeTemplatesImporting import *
from .templates.CodeTemplatesPrinting import *
from .templates.CodeTemplatesBranches import *
from .templates.CodeTemplatesTuples import *
from .templates.CodeTemplatesLists import *
from .templates.CodeTemplatesDicts import *
from .templates.CodeTemplatesCalls import *
from .templates.CodeTemplatesLoops import *
from .templates.CodeTemplatesSets import *

from .templates.CodeTemplatesExecEval import *

def enableDebug():
    templates = dict( globals() )

    class TemplateWrapper:
        """ Wrapper around templates, to better trace and control template usage. """
        def __init__( self, name, value ):
            self.name = name
            self.value = value

        def __str__( self ):
            return self.value

        def __mod__( self, other ):
            assert type( other ) is dict, self.name

            for key in other.keys():
                if "%%(%s)" % key not in self.value:
                    from logging import warning

                    warning( "Extra value '%s' provided to template '%s'.", key, self.name )

            return self.value % other

        def split( self, sep ):
            return self.value.split( sep )

    from nuitka.__past__ import iterItems

    for template_name, template_value in iterItems( templates ):
        if template_name.startswith( "_" ):
            continue

        if type( template_value ) is str:
            globals()[ template_name ] = TemplateWrapper(
                template_name,
                template_value
            )

from nuitka.Options import isDebug

if isDebug():
    enableDebug()
